import asyncio
import atexit
import json
import logging.config
import random
import re
import sys
import time
from pathlib import Path
from urllib.parse import quote, urlencode, parse_qs, urlsplit, urlunsplit

import aiohttp
import requests

from .config.log import log_config
from .config.settings import search_config

IN_PATH = Path('~/data/raw').expanduser()
OUT_PATH = Path(f'~/data/processed/search_results_{time.time_ns()}.json').expanduser()

reset = '\u001b[0m'
colors = [f'\u001b[{i}m' for i in range(30, 38)]
logging.config.dictConfig(log_config)
logger = logging.getLogger(__name__)


try:
    if get_ipython().__class__.__name__ == 'ZMQInteractiveShell':
        import nest_asyncio
        nest_asyncio.apply()
except:
    ...

if sys.platform != 'win32':
    try:
        import uvloop
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    except:
        ...
else:
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


def search(*args, out: str = 'data'):
    out_path = make_output_dirs(out)
    return asyncio.run(process(args, search_config, out_path))


async def process(queries: tuple, config: dict, out: Path) -> tuple:
    conn = aiohttp.TCPConnector(limit=len(queries), ssl=False)
    async with aiohttp.ClientSession(headers=__get_headers(), connector=conn) as s:
        return await asyncio.gather(*(paginate(q, s, config, out) for q in queries))


async def paginate(query: str, session: aiohttp.ClientSession, config: dict, out: Path) -> list[dict]:
    api = 'https://api.twitter.com/2/search/adaptive.json?'
    config['q'] = query
    data, next_cursor = await backoff(lambda: get(session, api, config), query)
    all_data = []
    c = colors.pop() if colors else ''
    while next_cursor:
        logger.debug(f'{c}{query}{reset}')
        config['cursor'] = next_cursor
        data, next_cursor = await backoff(lambda: get(session, api, config), query)
        data['query'] = query
        (out / f'raw/{time.time_ns()}.json').write_text(json.dumps(data, indent=4))
        all_data.append(data)
    return all_data


async def backoff(fn, info, retries=12):
    for i in range(retries + 1):
        try:
            data, next_cursor = await fn()
            if not data.get('globalObjects', {}).get('tweets'):
                raise Exception
            return data, next_cursor
        except Exception as e:
            if i == retries:
                logger.debug(f'Max retries exceeded\n{e}')
                return
            t = 2 ** i + random.random()
            logger.debug(f'No data for: \u001b[1m{info}\u001b[0m | retrying in {f"{t:.2f}"} seconds\t\t{e}')
            time.sleep(t)


async def get(session: aiohttp.ClientSession, api: str, params: dict) -> tuple[dict, str]:
    url = set_qs(api, params, update=True)
    r = await session.get(url)
    data = await r.json()
    next_cursor = get_cursor(data)
    return data, next_cursor


def get_cursor(res: dict):
    try:
        for instr in res['timeline']['instructions']:
            if replaceEntry := instr.get('replaceEntry'):
                cursor = replaceEntry['entry']['content']['operation']['cursor']
                if cursor['cursorType'] == 'Bottom':
                    return cursor['value']
                continue
            for entry in instr['addEntries']['entries']:
                if entry['entryId'] == 'cursor-bottom-0':
                    return entry['content']['operation']['cursor']['value']
    except Exception as e:
        logger.debug(e)


def set_qs(url: str, qs: dict, update=False) -> str:
    *_, q, f = urlsplit(url)
    return urlunsplit((*_, urlencode(qs | parse_qs(q) if update else qs, doseq=True, quote_via=quote, safe='()'), f))


def __get_headers(fname: str = None) -> dict:
    if fname:
        with open(fname) as fp:
            return {y.group(): z.group()
                    for x in fp.read().splitlines()
                    if (y := re.search('^[\w-]+(?=:\s)', x),
                        z := re.search(f'(?<={y.group()}:\s).*', x))}
    # default
    headers = {
        'authorization': 'Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs=1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36'
    }
    r = requests.post('https://api.twitter.com/1.1/guest/activate.json', headers=headers)
    headers['x-guest-token'] = r.json()['guest_token']
    return headers


def make_output_dirs(path: str) -> Path:
    p = Path(f'~/{path}').expanduser()
    (p / 'raw').mkdir(parents=True, exist_ok=True)
    (p / 'processed').mkdir(parents=True, exist_ok=True)
    (p / 'final').mkdir(parents=True, exist_ok=True)
    return p


@atexit.register
def combine_results(in_path: Path = IN_PATH, out_path: Path = OUT_PATH):
    try:
        out_path.write_text(json.dumps({
            k: v
            for p in in_path.iterdir() if p.suffix == '.json'
            for k, v in json.loads(p.read_text())['globalObjects']['tweets'].items()
        }, indent=2))
    except Exception as e:
        logger.debug(f'FAILED to combine search results, {e}')



import sys
from requests import Session
from .constants import SUCCESS, WARN, BOLD, RESET


def update_token(session: Session, key: str, url: str, payload: dict) -> Session:
    try:
        headers = {
            "authorization": 'Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs%3D1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA',
            "content-type": "application/json",
            "user-agent": 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.114 Safari/537.36',
            "x-guest-token": session.cookies.get('guest_token'),
            "x-csrf-token": session.cookies.get("ct0"),
            "x-twitter-auth-type": "OAuth2Session" if session.cookies.get("auth_token") else '',
            "x-twitter-active-user": "yes",
            "x-twitter-client-language": 'en',
        }
        r = session.post(url, headers=headers, json=payload).json()
        # print(f'{SUCCESS}{sys._getframe(1).f_code.co_name}{RESET}')
        session.cookies.set(key, r[key])
    except KeyError as e:
        print(f'[{WARN}FAILED{RESET}] failed to update token at {BOLD}{sys._getframe(1).f_code.co_name}{RESET}')
    return session


def init_guest_token(session: Session) -> Session:
    return update_token(session, 'guest_token', 'https://api.twitter.com/1.1/guest/activate.json', {})


def flow_start(session: Session) -> Session:
    return update_token(session, 'flow_token', 'https://api.twitter.com/1.1/onboarding/task.json?flow_name=login', {
        "input_flow_data": {
            "flow_context": {"debug_overrides": {}, "start_location": {"location": "splash_screen"}}
        }, "subtask_versions": {}
    })


def flow_instrumentation(session: Session) -> Session:
    return update_token(session, 'flow_token', 'https://api.twitter.com/1.1/onboarding/task.json', {
        "flow_token": session.cookies.get('flow_token'),
        "subtask_inputs": [{
            "subtask_id": "LoginJsInstrumentationSubtask",
            "js_instrumentation": {"response": "{}", "link": "next_link"}
        }],
    })


def flow_username(session: Session) -> Session:
    return update_token(session, 'flow_token', 'https://api.twitter.com/1.1/onboarding/task.json', {
        "flow_token": session.cookies.get('flow_token'),
        "subtask_inputs": [{
            "subtask_id": "LoginEnterUserIdentifierSSO",
            "settings_list": {
                "setting_responses": [{
                    "key": "user_identifier",
                    "response_data": {"text_data": {"result": session.cookies.get('username')}}
                }], "link": "next_link"}}],
    })


def flow_password(session: Session) -> Session:
    return update_token(session, 'flow_token', 'https://api.twitter.com/1.1/onboarding/task.json', {
        "flow_token": session.cookies.get('flow_token'),
        "subtask_inputs": [{
            "subtask_id": "LoginEnterPassword",
            "enter_password": {"password": session.cookies.get('password'), "link": "next_link"}}]
    })


def flow_duplication_check(session: Session) -> Session:
    return update_token(session, 'flow_token', 'https://api.twitter.com/1.1/onboarding/task.json', {
        "flow_token": session.cookies.get('flow_token'),
        "subtask_inputs": [{
            "subtask_id": "AccountDuplicationCheck",
            "check_logged_in_account": {"link": "AccountDuplicationCheck_false"},
        }],
    })


def execute_login_flow(session: Session) -> Session:
    session = init_guest_token(session)
    for fn in [flow_start, flow_instrumentation, flow_username, flow_password, flow_duplication_check]:
        session = fn(session)
    return session


def login(username: str, password: str) -> Session:
    session = Session()
    session.cookies.update({
        "username": username,
        "password": password,
        "guest_token": None,
        "flow_token": None,
    })
    session = execute_login_flow(session)
    print(f'[{SUCCESS}SUCCESS{RESET}] {BOLD}{username}{RESET} logged in successfully')
    return session



account_settings = {
    "address_book_live_sync_enabled": False,
    "allow_ads_personalization": False,
    "allow_authenticated_periscope_requests": True,
    "allow_dm_groups_from": "following",
    "allow_dms_from": "following",  # all
    "allow_location_history_personalization": False,
    "allow_logged_out_device_personalization": False,
    "allow_media_tagging": "none",  # all, following
    "allow_sharing_data_for_third_party_personalization": False,
    "alt_text_compose_enabled": None,
    "always_use_https": True,
    "autoplay_disabled": False,
    "country_code": "us",
    "discoverable_by_email": False,
    "discoverable_by_mobile_phone": False,
    "display_sensitive_media": True,
    "dm_quality_filter": "enabled",  # disabled
    "dm_receipt_setting": "all_disabled",  # all_enabled
    "geo_enabled": False,
    "include_alt_text_compose": True,
    "include_mention_filter": True,
    "include_nsfw_admin_flag": True,
    "include_nsfw_user_flag": True,
    "include_ranked_timeline": True,
    "language": "en",
    "mention_filter": "unfiltered",
    "nsfw_admin": False,
    "nsfw_user": False,
    "personalized_trends": True,
    "protected": False,
    "ranked_timeline_eligible": None,
    "ranked_timeline_setting": None,
    "require_password_login": False,
    "requires_login_verification": False,
    "settings_metadata": {},
    "sleep_time": {
        "enabled": False,
        "end_time": None,
        "start_time": None
    },
    "translator_type": "none",
    "universal_quality_filtering_enabled": "enabled",
    "use_cookie_personalization": False,
    ## todo: not yet implemented - requires additional steps
    # "allow_contributor_request": "all",
    # "protect_password_reset": False,
}
notification_settings = {
    "cursor": "-1",
    "include_profile_interstitial_type": "1",
    "include_blocking": "1",
    "include_blocked_by": "1",
    "include_followed_by": "1",
    "include_want_retweets": "1",
    "include_mute_edge": "1",
    "include_can_dm": "1",
    "include_can_media_tag": "1",
    "include_ext_has_nft_avatar": "1",
    "include_ext_is_blue_verified": "1",
    "include_ext_verified_type": "1",
    "skip_status": "1",
}

follow_settings = {
    "include_profile_interstitial_type": "1",
    "include_blocking": "1",
    "include_blocked_by": "1",
    "include_followed_by": "1",
    "include_want_retweets": "1",
    "include_mute_edge": "1",
    "include_can_dm": "1",
    "include_can_media_tag": "1",
    "include_ext_has_nft_avatar": "1",
    "include_ext_is_blue_verified": "1",
    "include_ext_verified_type": "1",
    "skip_status": "1",
}

account_search_settings = {
    "optInFiltering": True,  # filter out nsfw content
    "optInBlocking": True,  # filter out blocked accounts
}

profile_settings = {
    "birthdate_day": int,
    "birthdate_month": int,
    "birthdate_year": int,  # 1985
    "birthdate_visibility": str,  # "self",
    "birthdate_year_visibility": str,  # "self",
    "displayNameMaxLength": int,  # "50",
    "url": str,  # "https://example.com",
    "name": str,  # "foo",
    "description": str,  # "bar",
    "location": str,  # "world",
}

search_config = {
    "include_profile_interstitial_type": 1,
    "include_blocking": 1,
    "include_blocked_by": 1,
    "include_followed_by": 1,
    "include_want_retweets": 1,
    "include_mute_edge": 1,
    "include_can_dm": 1,
    "include_can_media_tag": 1,
    "include_ext_has_nft_avatar": 1,
    "include_ext_is_blue_verified": 1,
    "include_ext_verified_type": 1,
    "skip_status": 1,
    "cards_platform": "Web-12",
    "include_cards": 1,
    "include_ext_alt_text": "true",
    "include_ext_limited_action_results": "false",
    "include_quote_count": "true",
    "include_reply_count": 1,
    "tweet_mode": "extended",
    "include_ext_collab_control": "true",
    "include_ext_views": "true",
    "include_entities": "true",
    "include_user_entities": "true",
    "include_ext_media_color": "true",
    "include_ext_media_availability": "true",
    "include_ext_sensitive_media_warning": "true",
    "include_ext_trusted_friends_metadata": "true",
    "send_error_codes": "true",
    "simple_quoted_tweet": "true",
    "query_source": "typed_query",
    "count": 100,
    "q": "",
    "requestContext": "launch",
    "pc": 1,
    "spelling_corrections": 1,
    "include_ext_edit_control": "true",
    "ext": "mediaStats,highlightedLabel,hasNftAvatar,voiceInfo,birdwatchPivot,enrichments,superFollowMetadata,unmentionInfo,editControl,collab_control,vibe"
}



import asyncio
import logging.config
import math
import random
import re
import sys
import time
from copy import deepcopy
from pathlib import Path
from urllib.parse import urlsplit

import ujson
from aiohttp import ClientSession, TCPConnector

from .config.operations import operations
from .config.log import log_config
from .constants import *
from .login import Session
from .utils import find_key, build_query, get_headers

try:
    if get_ipython().__class__.__name__ == 'ZMQInteractiveShell':
        import nest_asyncio

        nest_asyncio.apply()
except:
    ...

if sys.platform != 'win32':
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
else:
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

ID = 'ID'
DUP_LIMIT = 5
logging.config.dictConfig(log_config)
logger = logging.getLogger(__name__)


def tweets(session: Session, ids: list[int], limit=math.inf):
    return run(session, ids, Operation.Data.UserTweets, limit)


def tweets_and_replies(session: Session, ids: list[int], limit=math.inf):
    return run(session, ids, Operation.Data.UserTweetsAndReplies, limit)


def likes(session: Session, ids: list[int], limit=math.inf):
    return run(session, ids, Operation.Data.Likes, limit)


def media(session: Session, ids: list[int], limit=math.inf):
    return run(session, ids, Operation.Data.UserMedia, limit)


def followers(session: Session, ids: list[int], limit=math.inf):
    return run(session, ids, Operation.Data.Followers, limit)


def following(session: Session, ids: list[int], limit=math.inf):
    return run(session, ids, Operation.Data.Following, limit)


def favoriters(session: Session, ids: list[int], limit=math.inf):
    return run(session, ids, Operation.Data.Favoriters, limit)


def retweeters(session: Session, ids: list[int], limit=math.inf):
    return run(session, ids, Operation.Data.Retweeters, limit)


def tweets_details(session: Session, ids: list[int], limit=math.inf):
    return run(session, ids, Operation.Data.TweetDetail, limit)


# no pagination needed
def tweets_by_rest_id(session: Session, ids: list[int]):
    return run(session, ids, Operation.Data.TweetResultByRestId)


# no pagination needed
def user_by_screen_name(session: Session, ids: list[str]):
    return run(session, ids, Operation.Data.UserByScreenName)


# no pagination needed
def user_by_rest_id(session: Session, ids: list[int]):
    return run(session, ids, Operation.Data.UserByRestId)


# no pagination needed (special batch query)
def users_by_rest_ids(session: Session, ids: list[int]):
    name, key = Operation.Data.UsersByRestIds
    params = deepcopy(operations[name])
    qid = params['queryId']
    params['variables']['userIds'] = ids
    q = build_query(params)
    url = f"https://api.twitter.com/graphql/{qid}/{name}?{q}"
    headers = get_headers(session)
    headers['content-type'] = "application/json"
    users = session.get(url, headers=headers).json()
    return users


def run(session: Session, ids: list[str | int], operation: tuple, limit=None):
    res = query(session, ids, operation)
    if limit is None:
        return res
    return asyncio.run(pagination(session, res, operation, limit))


def query(session: Session, ids: list[any], operation: tuple) -> list:
    name, key = operation
    params = deepcopy(operations[name])
    qid = params['queryId']
    urls = []
    for _id in ids:
        params['variables'][key] = _id
        q = build_query(params)
        urls.append((_id, f"https://api.twitter.com/graphql/{qid}/{name}?{q}"))
    headers = get_headers(session)
    headers['content-type'] = "application/json"
    res = asyncio.run(process(session, urls, headers))
    save_data(res, name)
    return res


async def process(session: Session, urls: list, headers: dict) -> tuple:
    conn = TCPConnector(limit=0, ssl=False, ttl_dns_cache=69)
    async with ClientSession(headers=headers, connector=conn) as s:
        # add cookies from logged-in session
        s.cookie_jar.update_cookies(session.cookies)
        return await asyncio.gather(*(get(s, u) for u in urls))


async def get(session: ClientSession, url: tuple) -> dict:
    identifier, api_url = url
    logger.debug(f'processing: {url}')
    try:
        r = await session.get(api_url)
        limits = {k: v for k, v in r.headers.items() if 'x-rate-limit' in k}
        logger.debug(f'{limits = }')
        if r.status == 429:
            logger.debug(f'rate limit exceeded: {url}')
            return {}
        data = await r.json()
        return {ID: identifier, **data}
    except Exception as e:
        logger.debug(f'failed to download {url}: {e}')


async def pagination(session: Session, res: list, operation: tuple, limit: int) -> tuple:
    conn = TCPConnector(limit=len(res), ssl=False, ttl_dns_cache=69)
    headers = get_headers(session)
    headers['content-type'] = "application/json"
    async with ClientSession(headers=headers, connector=conn) as s:
        # add cookies from logged-in session
        s.cookie_jar.update_cookies(session.cookies)
        return await asyncio.gather(*(paginate(s, data, operation, limit) for data in res))


async def paginate(session: ClientSession, data: dict, operation: tuple, limit: int):
    def get_cursor(data):
        # inefficient, but need to deal with arbitrary schema
        entries = find_key(data, 'entries')
        if entries:
            for entry in entries.pop():
                if entry.get('entryId', '').startswith('cursor-bottom'):
                    content = entry['content']
                    if itemContent := content.get('itemContent'):
                        return itemContent['value']  # v2 cursor
                    return content['value']  # v1 cursor

    all_data = []
    try:
        name, key = operation
        params = deepcopy(operations[name])
        qid = params['queryId']

        ids = set()
        counts = []

        params['variables'][key] = data[ID]
        cursor = get_cursor(data)

        while 1:
            params['variables']['cursor'] = cursor
            query = build_query(params)
            url = f"https://api.twitter.com/graphql/{qid}/{name}?{query}"

            # code [353]: "This request requires a matching csrf cookie and header."
            r, _data = await backoff(lambda: session.get(url))
            if csrf := r.cookies.get("ct0"):
                session.headers.update({"x-csrf-token": csrf.value})
            session.cookie_jar.update_cookies(r.cookies)

            tagged_data = _data | {ID: data[ID]}
            save_data([tagged_data], name)
            all_data.append(tagged_data)
            cursor = get_cursor(_data)
            logger.debug(f'{cursor = }')
            ids |= set(find_key(tagged_data, 'rest_id'))
            logger.debug(f'({data[ID]})\t{len(ids)} unique results')
            counts.append(len(ids))

            success_message = f'[{SUCCESS}SUCCESS{RESET}] done pagination'
            # followers/following have "0|"
            if not cursor or cursor.startswith('0|'):
                logger.debug(f'{success_message}\tlast cursor: {cursor}')
                break
            if len(ids) >= limit:
                logger.debug(f'{success_message}\tsurpassed limit of {limit} results')
                break
            # did last 5 requests return duplicate data?
            if len(counts) > DUP_LIMIT and len(set(counts[-1:-DUP_LIMIT:-1])) == 1:
                logger.debug(f'{success_message}\tpast {DUP_LIMIT} requests returned duplicate data')
                break
    except Exception as e:
        logger.debug(f'paginate falied: {e}')
    # save_data(all_data, name)
    return all_data


async def backoff(fn, retries=12):
    for i in range(retries + 1):
        try:
            r = await fn()
            data = await r.json()
            return r, data
        except Exception as e:
            if i == retries:
                logger.debug(f'{WARN}Max retries exceeded{RESET}\n{e}')
                return
            t = 2 ** i + random.random()
            logger.debug(f'retrying in {f"{t:.2f}"} seconds\t\t{e}')
            time.sleep(t)


def save_data(data: list, name: str = ''):
    try:
        for d in data:
            path = Path(f'data/raw/{d[ID]}')
            path.mkdir(parents=True, exist_ok=True)
            with open(path / f'{time.time_ns()}_{name}.json', 'w') as fp:
                ujson.dump(d, fp, indent=4)
    except KeyError as e:
        logger.debug(f'failed to save data: {e}')


def download(session: Session, post_url: str, cdn_url: str, path: str = 'media', chunk_size: int = 4096) -> None:
    """
    Download file

    @param post_url: the actual post url
    @param cdn_url: the cdn url
    @param path: path to save media
    @param chunk_size: chunk size in bytes
    @return: None
    """
    Path(path).mkdir(parents=True, exist_ok=True)
    name = urlsplit(post_url).path.replace('/', '_')[1:]
    ext = urlsplit(cdn_url).path.split('/')[-1]
    try:
        with open(f'{path}/{name}_{ext}', 'wb') as fp:
            r = session.get(cdn_url, stream=True)
            for chunk in r.iter_content(chunk_size=chunk_size):
                fp.write(chunk)
    except Exception as e:
        logger.debug(f'FAILED to download video: {post_url} {e}')


def download_media(session: Session, ids: list[int], photos: bool = True, videos: bool = True) -> None:
    res = tweets_by_rest_id(session, ids)
    for r in res:
        user_id = find_key(r, 'user_results')[0]['result']['rest_id']
        url = f'https://twitter.com/{user_id}/status/{r[ID]}'  # evaluates to username in browser
        media = [y for x in find_key(r, 'media') for y in x]  # in case of arbitrary schema
        if photos:
            photos = list({u for m in media if 'ext_tw_video_thumb' not in (u := m['media_url_https'])})
            logger.debug(f'{photos = }')
            if photos:
                [download(session, url, photo) for photo in photos]
        if videos:
            videos = [x['variants'] for m in media if (x := m.get('video_info'))]
            hq_videos = {sorted(v, key=lambda d: d.get('bitrate', 0))[-1]['url'] for v in videos}
            logger.debug(f'{videos = }')
            logger.debug(f'{hq_videos = }')
            if hq_videos:
                [download(session, url, video) for video in hq_videos]



log_config = {
    "version": 1,
    "formatters": {
        "simple": {
            "format": "%(asctime)s.%(msecs)03d %(levelname)s: %(message)s"
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "simple",
            "stream": "ext://sys.stdout"
        },
        "file": {
            "class": "logging.FileHandler",
            "level": "DEBUG",
            "formatter": "simple",
            "filename": "debug.log",
            "encoding": "utf8",
            "mode": "a"
        }
    },
    "loggers": {
        "myLogger": {
            "level": "DEBUG",
            "handlers": [
                "console"
            ],
            "propagate": "no"
        }
    },
    "root": {
        "level": "DEBUG",
        "handlers": [
            "console",
            "file"
        ]
    }
}



import sys
from textwrap import dedent
from setuptools import find_packages, setup

install_requires = [
    "tqdm",
    "ujson",
    "nest_asyncio",
    "aiohttp",
    "requests",
]
if sys.platform != 'win32':
    install_requires.append('uvloop')

setup(
    name="twitter-api-client",
    version="0.3.9",
    description="Twitter API",
    long_description=dedent('''
    Complete implementation of the undocumented Twitter API
    
    Includes tools to **scrape**, **automate**, and **search** twitter
    
    ### Installation
    
   ```python
    from twitter.account import Account
    from twitter.login import login
    
    username,password = ...,...
    account = Account(login(username,password))
    
    account.create_poll('test poll 123', ['hello', 'world', 'foo', 'bar'], 10080)
    
    # DM 1 user
    account.dm([123], 'hello world', filename='test.png')
    
    # DM group of users
    account.dm([123,234,345], 'foo bar', filename='test.mp4')
    
    # schedule a tweet (date str or timestamp)
    account.schedule_tweet('scheduled hello', 1679912795, media=['test.jpg'])
    
    # schedule a reply tweet (date str or timestamp)
    account.schedule_tweet('scheduled world', '2023-03-25 19:11', media=['test.jpg'], reply_to=645)
    
    account.unschedule_tweet(321)
    
    # tweets
    account.tweet('test 123')
    account.tweet('test 234', media=['test.mp4'])
    account.tweet('test 345', media=['test.jpg', 'test.png', 'test.jpeg','test.jfif'])
    account.tweet('test 456', media=[{'file': 'test.jpeg', 'tagged_users': [123234345456], 'alt': 'some image'}])
    account.untweet(123)
    account.retweet(1633609779745820675)
    account.unretweet(1633609779745820675)
    account.quote(1633609779745820675, 'elonmusk', 'test 123')
    account.reply(1633609779745820675, 'test 123')
    account.like(1633609779745820675)
    account.unlike(1633609779745820675)
    account.bookmark(1633609779745820675)
    account.unbookmark(1633609779745820675)
    account.pin(1635479755364651008)
    account.unpin(1635479755364651008)
    
    # users
    account.follow(50393960)
    account.unfollow(50393960)
    account.mute(50393960)
    account.unmute(50393960)
    account.enable_notifications(50393960)
    account.disable_notifications(50393960)
    account.block(50393960)
    account.unblock(50393960)
    
    # other
    account.stats(50393960)
    
    # user profile
    account.update_profile_image('test.jpg')
    account.update_profile_banner('test.png')
    account.update_profile_info(name='Foo Bar', description='test 123', location='Victoria, BC')
    
    # topics
    account.follow_topic(808713037230157824)
    account.unfollow_topic(808713037230157824)
    
    # lists
    account.create_list('My List', 'description of my list', private=False)
    account.update_list(543, 'My Updated List', 'some updated description', private=False)
    account.update_list_banner(543, 'test.png')
    account.delete_list_banner(543)
    account.add_list_member(543, 50393960)
    account.remove_list_member(543, 50393960)
    account.delete_list(543)
    account.pin_list(543)
    account.unpin_list(543)
    
    # refresh all pinned lists in this order
    account.update_pinned_lists([543,432,321])
    
    # unpin all lists
    account.update_pinned_lists([])
    
    # example configuration
    account.update_settings({
        "address_book_live_sync_enabled": False,
        "allow_ads_personalization": False,
        "allow_authenticated_periscope_requests": True,
        "allow_dm_groups_from": "following",
        "allow_dms_from": "following",
        "allow_location_history_personalization": False,
        "allow_logged_out_device_personalization": False,
        "allow_media_tagging": "none",
        "allow_sharing_data_for_third_party_personalization": False,
        "alt_text_compose_enabled": None,
        "always_use_https": True,
        "autoplay_disabled": False,
        "country_code": "us",
        "discoverable_by_email": False,
        "discoverable_by_mobile_phone": False,
        "display_sensitive_media": False,
        "dm_quality_filter": "enabled",
        "dm_receipt_setting": "all_disabled",
        "geo_enabled": False,
        "include_alt_text_compose": True,
        "include_mention_filter": True,
        "include_nsfw_admin_flag": True,
        "include_nsfw_user_flag": True,
        "include_ranked_timeline": True,
        "language": "en",
        "mention_filter": "unfiltered",
        "nsfw_admin": False,
        "nsfw_user": False,
        "personalized_trends": True,
        "protected": False,
        "ranked_timeline_eligible": None,
        "ranked_timeline_setting": None,
        "require_password_login": False,
        "requires_login_verification": False,
        "sleep_time": {
            "enabled": False,
            "end_time": None,
            "start_time": None
        },
        "translator_type": "none",
        "universal_quality_filtering_enabled": "enabled",
        "use_cookie_personalization": False,
    })
    
    # example configuration
    account.update_search_settings({
        "optInFiltering": True,  # filter out nsfw content
        "optInBlocking": True,  # filter out blocked accounts
    })
    
    ## account.change_password('old password', 'new password')
    ## account.logout_all_sessions()
    ```
    
    
    ### Scraping
    
    #### Get all user/tweet data
    
    ```python
    from twitter import scraper
    from twitter.login import login
    
    username, password = ..., ...
    s = login(username, password)  # session
    
    ####### User Data ########
    users = scraper.user_by_screen_name(s, ['bob123', 'jim456', 'stanley789'])
    tweets = scraper.tweets(s, [123, 234, 345])
    likes = scraper.likes(s, [123, 234, 345])
    tweets_and_replies = scraper.tweets_and_replies(s, [123, 234, 345])
    media = scraper.media(s, [123, 234, 345])
    following = scraper.following(s, [123, 234, 345])
    followers = scraper.followers(s, [123, 234, 345])
    
    ######## Tweet Data ########
    tweets_by_ids = scraper.tweets_by_rest_id(s, [456, 567, 678])
    tweets_details = scraper.tweets_details(s, [456, 567, 678])
    retweeters = scraper.retweeters(s, [456, 567, 678])
    favoriters = scraper.favoriters(s, [456, 567, 678])
    
    scraper.download_media(s, [456, 567, 678])
    ```
    
    #### Most recent ~100 results of user/tweet data
    
    ```python
    from twitter.login import login
    from twitter.constants import Operation
    from twitter import scraper
    from functools import partial
    
    username, password = ..., ...
    session = login(username, password)
    
    user_ids = [123, 234, 345, 456]
    user_query = partial(scraper.query, session, user_ids)
    
    tweets = user_query(Operation.Data.UserTweets)
    likes = user_query(Operation.Data.Likes)
    followers = user_query(Operation.Data.Followers)
    ```
    
    #### Search
    
    ```python   
    from twitter.search import search
    
    search(
        '(#dogs OR #cats) min_retweets:500',
        'min_faves:10000 @elonmusk until:2023-02-16 since:2023-02-01',
        'brasil portugal -argentina',
        'paperswithcode -tensorflow -tf',
        'skateboarding baseball guitar',
        'cheese bread butter',
        'ios android',
    )
    ```
    '''),
    long_description_content_type='text/markdown',
    author="Trevor Hobenshield",
    author_email="trevorhobenshield@gmail.com",
    url="https://github.com/trevorhobenshield/twitter-api",
    install_requires=install_requires,
    keywords="twitter api client async search automation bot scrape",
    packages=find_packages(),
    include_package_data=True,
)
